namespace Classes
{
    interface IDisplayable
    {
        void Display();
    }
    interface IMovieInfo
    {
        void ShowMovieDetails();
        int GetDuration();
    }
    enum MovieGenre
    {
        None = 0,
        Action = 1,
        Comedy = 2,
        Drama = 4,
        Horror = 8,
        SciFi = 16,
        Romance = 32
    }
    public class MovieDetails
    {
        public required string Director { get; set; }
        public required int Duration { get; set; }
    }

    abstract class Media
    {
        static int _totalMediaCount = 0;
        static readonly string _defaultDescription;

        static Media()
        {
            Console.WriteLine("Статичний конструктор Media");
            _defaultDescription = "Немає опису";
        }

        public string Name { get; }
        public string Description { get; }
        protected Media(string name) : this(name, _defaultDescription){}
        protected Media(string name, string description)
        {
            Name = name;
            Description = description;
            _totalMediaCount++;
        }
        public virtual void Display()
        {
            Console.WriteLine($"Медіа: {Name} відображається...");
        }
        public abstract void ShowInfo();

        public static int GetTotalCount() => _totalMediaCount;

        public override bool Equals(object? obj)
        {
            if (obj is not Media other)
                return false;
            return Name == other.Name;
        }

        public override int GetHashCode() => Name.GetHashCode();

        public override string ToString() => $"Медіа: {Name}";
    }

    class Movie : Media, IMovieInfo
    {
        static readonly string _defaultDirector = "John Doe";
        
        static int _totalMovieCount;

        static Movie()
        {
            Console.WriteLine("Статичний конструктор Movie");
            _totalMovieCount = 0;
        }

        // Поле з ініціалізацією при оголошенні
        readonly MovieDetails details = new() { Director = _defaultDirector, Duration = 120 };

        public int Year { get; }
        public MovieGenre Genres { get; set; } = MovieGenre.None;

        public Movie(string name, int year) : base(name)
        {
            Console.WriteLine("Базовий конструктор Movie");
            Year = year;
            _totalMovieCount++;
        }

        public Movie(string name, int year, string director, int duration) 
            : base(name)
        {
            Console.WriteLine("Перевантажений конструктор Movie (base)");
            details.Director = director;
            details.Duration = duration;
            Year = year;
            _totalMovieCount++;
        }

        public Movie(string name, int year, string description) 
            : base(name, description)
        {
            Console.WriteLine("Перевантажений конструктор Movie (base)");
            Year = year;
            _totalMovieCount++;
        }
        public Movie(string name) : this(name, 0, "Невідомо", 120){}
        public override void ShowInfo()
        {
            base.Display();
            var genreList = Enum.GetValues<MovieGenre>()
                .Where(g => g != MovieGenre.None && (Genres & g) == g)
                .ToList();
            
            Console.WriteLine(
                $"Фільм: {Name}\n" +
                $"Рік: {(Year != 0 ? Year.ToString() : "Невідомо")}\n" +
                $"Жанри: {(genreList.Count != 0 ? string.Join(", ", genreList) : "Не вказано")}\n" +
                $"Тривалість: {GetDuration()} хв"
            );
        }

        public void ShowMovieDetails()
        {
            Console.WriteLine($"Режисер: {details.Director}, тривалість: {details.Duration} хв");
        }

        public int GetDuration() => details.Duration;

        public static int GetMovieCount() => _totalMovieCount;

        // Публічний метод (доступний з будь-якої частини програми)
        public static void PublicMethod()
        {
            Console.WriteLine("Це публічний метод");
        }

        // Приватний метод (доступний тільки всередині класу)
        private static void PrivateMethod()
        {
            Console.WriteLine("Це приватний метод");
        }

        // Захищений метод (доступний в класах-нащадках)
        protected static void ProtectedMethod()
        {
            Console.WriteLine("Це захищений метод");
        }

        // Внутрішній метод (доступний тільки в межах того ж самого збірки)
        internal static void InternalMethod()
        {
            Console.WriteLine("Це внутрішній метод");
        }

        // Метод, доступний в межах збірки та через нащадків (protected internal)
        protected internal static void ProtectedInternalMethod()
        {
            Console.WriteLine("Це protected internal метод");
        }
        private class Review(string author, string content)
        {
            public string Author { get; } = author;
            public string Content { get; } = content;

            public void ShowReview() => Console.WriteLine($"Відгук від {Author}: {Content}");
        }

        Review movieReview = new("Критик", "Дуже цікавий фільм!");

        public void ShowReview() => movieReview.ShowReview();
            public static void ShowGenreOperations()
        {
            MovieGenre genre1 = MovieGenre.Action | MovieGenre.Comedy;
            MovieGenre genre2 = MovieGenre.Drama | MovieGenre.Horror;
            
            Console.WriteLine($"Жанр 1: {genre1}");
            Console.WriteLine($"Жанр 2: {genre2}");
            
            Console.WriteLine($"Об'єднання (|): {genre1 | genre2}");
            Console.WriteLine($"Перетин (&): {genre1 & genre2}");
            Console.WriteLine($"Різниця (^): {genre1 ^ genre2}");
            Console.WriteLine($"Чи є Action у genre1? {(genre1 & MovieGenre.Action) != 0}");
            Console.WriteLine($"Чи є Horror у genre1? {(genre1 & MovieGenre.Horror) != 0}");
        }
        public static implicit operator string(Movie movie)
        {
            return movie.Name;
        }
        public static explicit operator int(Movie movie)
        {
            return movie.Year;
        }
    }
    class InitializationOrder
    { 
        static int _order = 1;
        static readonly int staticField = Print("Статичне поле при оголошенні", ref _order);
        readonly int instanceField = Print("Динамічне поле при оголошенні", ref _order);
        
        static InitializationOrder()
        {
            Print("Статичний конструктор", ref _order);
        }
        
        public InitializationOrder()
        {
            Print("Динамічний конструктор", ref _order);
        }
        
        private static int Print(string message, ref int order)
        {
            Console.WriteLine($"{order}. {message}");
            order++;
            return order;
        }
    }
        class MovieLight: IMovieInfo
    {
        static readonly string _defaultDirector = "John Doe";
        readonly MovieDetails details = new() { Director = _defaultDirector, Duration = 120 };



        static MovieLight()
        {
            Console.WriteLine("Статичний конструктор MovieLight");
        }

        public void ShowMovieDetails()
        {
            Console.WriteLine($"Режисер: {details.Director}, тривалість: {details.Duration} хв");
        }

        public int GetDuration() => details.Duration;
        public static void PublicMethod()
        {
            Console.WriteLine("Це публічний метод");
        }
        internal static void InternalMethod()
        {
            Console.WriteLine("Це внутрішній метод");
        }
        private class Review(string author, string content)
        {
            public string Author { get; } = author;
            public string Content { get; } = content;

            public void ShowReview() => Console.WriteLine($"Відгук від {Author}: {Content}");
        }
            public static void ShowGenreOperations()
        {
            MovieGenre genre1 = MovieGenre.Action | MovieGenre.Comedy;
            MovieGenre genre2 = MovieGenre.Drama | MovieGenre.Horror;
            
            Console.WriteLine($"Жанр 1: {genre1}");
            Console.WriteLine($"Жанр 2: {genre2}");
            
            Console.WriteLine($"Об'єднання (|): {genre1 | genre2}");
            Console.WriteLine($"Перетин (&): {genre1 & genre2}");
            Console.WriteLine($"Різниця (^): {genre1 ^ genre2}");
            Console.WriteLine($"Чи є Action у genre1? {(genre1 & MovieGenre.Action) != 0}");
            Console.WriteLine($"Чи є Horror у genre1? {(genre1 & MovieGenre.Horror) != 0}");
        }
    }
    class Program
    {
        static void DoubleValue(ref int value)
        {
            value *= 2;
            Console.WriteLine($"Значення після подвоєння: {value}");
        }

        static void InitializeAndDouble(out int value)
        {
            value = 10;
            value *= 2;
            Console.WriteLine($"Значення після ініціалізації і подвоєння: {value}");
        }

        static int DoubleValueStupid(int value) => value *= 2;
        static void Main()
        {
            Movie movie = new("Inception", 2010);

            //movie.PrivateMethod();// доступно тільки в класі Movie
            //movie.ProtectedMethod();// доступно в класах-нащадках
            Movie.InternalMethod();// доступно в межах збірки
            Movie.ProtectedInternalMethod();// доступно в межах збірки та через нащадків
            Movie.PublicMethod();// доступно всім і всюди
            movie.ShowReview();
            Movie.ShowGenreOperations();
            Movie unkownMovie = new("Страшне");
            unkownMovie.ShowInfo();
            _ = new InitializationOrder();
            Console.WriteLine("\nСтворюємо другий об'єкт\n");
            _ = new InitializationOrder();
            int numberRef = 5;
            Console.WriteLine($"Початкове значення: {numberRef}");
            DoubleValue(ref numberRef);
            Console.WriteLine($"Після виклику ref: {numberRef}\n");
            InitializeAndDouble(out int numberOut);
            Console.WriteLine($"Після виклику out: {numberOut}\n");
            int numberNormal = 5;
            int result = DoubleValueStupid(numberNormal);
            Console.WriteLine($"Після виклику без ref/out (оригінальне значення): {numberNormal}");
            Console.WriteLine($"Після виклику без ref/out (результат): {result}");
            object boxedValue = result;
            Console.WriteLine($"Boxing: {boxedValue} (тип: {boxedValue.GetType()})");
            int unboxedValue = (int)boxedValue;
            Console.WriteLine($"Unboxing: {unboxedValue} (тип: {unboxedValue.GetType()})");
            try
            {
                object wrongBox = 3.14;
                int wrongUnbox = (int)wrongBox;
            }
            catch (InvalidCastException e)
            {
                Console.WriteLine($"Помилка анбоксингу: {e.Message}");
            }
            string movieName = movie;
            int movieYear = (int)movie;
            Console.WriteLine($"Назва фільму: {movieName}, рік: {movieYear}");

            var stopwatch = new System.Diagnostics.Stopwatch();

            GC.Collect();
            GC.WaitForPendingFinalizers();

            long memoryBefore = GC.GetTotalMemory(true);

            var movies = new MovieLight[10_000_000];

            stopwatch.Start();
            for (long i = 0; i < movies.Length; i++)
            {
                movies[i] = new MovieLight();
            }
            stopwatch.Stop();
            Console.WriteLine($"Час створення об'єктів: {stopwatch.ElapsedMilliseconds} мс");

            GC.Collect();
            GC.WaitForPendingFinalizers();

            long memoryAfter = GC.GetTotalMemory(true);
            long memoryUsed = memoryAfter - memoryBefore;

            Console.WriteLine($"Використано памʼяті: {memoryUsed / 1024.0 / 1024.0:F2} МБ");

            stopwatch.Restart();
            for (long i = 0; i < movies.Length; i++)
            {
                _ = movies[i].GetHashCode();
            }
            stopwatch.Stop();
            Console.WriteLine($"Час отримання хеш-кодів: {stopwatch.ElapsedMilliseconds} мс");
        }
    }
}
